"""Util functions"""
import os
import sys
import zipfile
from pathlib import Path
from platform import system
from subprocess import Popen, PIPE, run

import yaml
from pylint.checkers.strings import arg_matches_format_type

from .classroom.model import Course, CourseWork, Submission, Student, Attachment


def capture_io(script_name: str, inputs: list) -> str:
    """
    :param script_name:  name of the script to be tested
    :param inputs: list of inputs, one element per line
    :return: returns script output for the test case
    """

    inputs = "\n".join(inputs)

    environ = os.environ.copy()
    environ['PYTHONIOENCODING'] = 'utf-8'

    p = Popen(
        [sys.executable, script_name],
        stdin=PIPE,
        stdout=PIPE,
        env=environ,
        stderr=PIPE,
        encoding="utf8")

    stdout, stderr = p.communicate(input=inputs)
    output_parts = stdout.split('\n')

    if output_parts and not output_parts[-1]:
        return '\n'.join(stdout.split('\n')[:-1])
    else:
        return '\n'.join(output_parts)


def extract_zip(source: str, destination: str) -> None:
    """
    Extract zip file at source to destination
    :param source:
    :param destination:
    :return:
    """
    with zipfile.ZipFile(source, 'r') as output:
        output.extractall(destination)


def create_unique_dir(parent_dir: str, new_dir_name: str) -> str:
    """
    creates a directory named new_dir_name
    if such a directory exists, it new directory will be created with unique name
    """
    os.chdir(parent_dir)

    if os.path.exists(new_dir_name) and os.listdir(new_dir_name):
        i = 1
        while os.path.exists(f"{new_dir_name}_{i}") and os.listdir(f"{new_dir_name}_{i}"):
            i += 1

        new_path = f"{new_dir_name}_{i}"

        os.mkdir(new_path)

        return new_path

    os.mkdir(new_dir_name)

    return new_dir_name


def get_current_course() -> Course:
    """
    Searches for .course.yaml file in current directory and returns the course entity from it
    :return:
    :raises: FileNotFoundError if .course.yaml file is not found
    """
    current_dir = [file for file in os.listdir('.') if os.path.isfile(file) and file == ".course.yaml"]
    if '.course.yaml' not in current_dir:
        raise FileNotFoundError("Can't find .course.yaml in this directory")

    with open('.course.yaml', 'r', encoding='utf8') as course_file:
        course = yaml.safe_load(course_file.read())

    return Course(id=course['id'], name=course['name'])


def find_descendant(file_name: str) -> str:
    """
    looks in previous paths for a file. returns a path to a file
    """

    path = Path(os.getcwd())

    while not os.path.exists(os.path.join(path, file_name)) and not os.path.ismount(path):
        path = path.parent.absolute()

    return_path = os.path.join(path, file_name)

    if not os.path.exists(return_path):
        raise FileNotFoundError(file_name)

    return os.path.join(path, file_name)


def write_course_yaml(course: Course, path: str) -> None:
    with open(os.path.join(path, ".course.yaml"), "w", encoding="utf-8") as f:
        yaml.dump(course.__dict__, f)


def read_course_yaml(path: str) -> Course:
    with open(os.path.join(path, ".course.yaml"), "r", encoding="utf-8") as f:
        course_info = yaml.safe_load(f)

    return Course(id=course_info["id"], name=course_info["name"], description=course_info["description"],
                  section=course_info["section"])


def write_course_work_yaml(course_work: CourseWork, path: str) -> None:
    with open(os.path.join(path, ".course_work.yaml"), "w", encoding="utf-8") as f:
        yaml.dump(course_work.__dict__, f)


def read_course_work_yaml(path: str) -> CourseWork:
    with open(os.path.join(path, ".course_work.yaml"), "r", encoding="utf-8") as f:
        course_work_info = yaml.safe_load(f)

    return CourseWork(id=course_work_info["id"],
                      title=course_work_info['title'],
                      description=course_work_info['description'],
                      course_id=course_work_info['course_id'],
                      link=course_work_info['link'])


def read_sumbission_yaml(path: str) -> Submission:
    with open(os.path.join(path, ".submission.yaml"), "r", encoding="utf-8") as f:
        submission_info = yaml.safe_load(f)

    return Submission(id=submission_info["id"], user_id=submission_info["user_id"],
                      course_id=submission_info["course_id"], coursework_id=submission_info["coursework_id"],
                      creation_time=submission_info["creation_time"], update_time=submission_info["update_time"],
                      attachments=[Attachment(id=value["id"], title=key, link=value["link"]) for key, value in
                                   submission_info["attachments"].items()], link=submission_info["link"])


def write_submission_yaml(submission: Submission, student: Student, path: str) -> None:
    with open(os.path.join(path, ".submission.yaml"), "w", encoding="utf-8") as f:
        submission_dict = submission.__dict__.copy()
        submission_dict["attachments"] = {attachment.title: {"id": attachment.id, "link": attachment.link} for
                                          attachment in submission_dict["attachments"]}
        submission_dict["student_name"] = student.name
        yaml.dump(submission_dict, f, allow_unicode=True)


def map_pylint_arguments(arguments_str: str) -> list[str]:
    """
    gets a string with parameters for pylint, returns a list of parametrs ready for usage
    """

    arguments_list = arguments_str.split(",")
    arguments_list = map(str.strip, arguments_list)
    arguments_list = filter(lambda arg: arg != '', arguments_list)

    return list(map(lambda arg: "--" + arg, arguments_list))



import argparse
import doctest
import importlib.util
import inspect
import os
import unittest
from typing import Type
from unittest import TestLoader, TestSuite, TextTestRunner

from pylint.lint import Run
from pylint.reporters.text import TextReporter
from six import StringIO

from .converter import create_unittest
from .globals import get_testing_script_path, set_testing_script_path
from .utils import map_pylint_arguments

PYLINT_WEIGHT = .25


class GraderTestResult(unittest.TextTestResult):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.total_points = None
        self.gained_points = None

        self.output = None


class GraderTestRunner(TextTestRunner):
    """Test runner that counts points for each successful test case."""

    def __init__(self, *args, pylint_args=[], **kwargs):
        super().__init__(*args, **kwargs)
        self.pylint_args = pylint_args

    def run_pylint(self) -> tuple[str, float]:
        """
        Run's pylint and returns how many points gained on scale from 0 to 1.
        """
        pylint_output = StringIO()
        reporter = TextReporter(pylint_output)
        Run(args=self.pylint_args + [get_testing_script_path()], reporter=reporter, exit=False)

        lint_message = pylint_output.getvalue()

        pep8_grade = 0
        for line in lint_message.splitlines():
            if "rated at" in line:
                try:
                    pep8_grade = float(line.split(" ")[6].split("/10")[0]) / 10
                    pep8_grade = pep8_grade if pep8_grade >= 0 else 0

                except:
                    pep8_grade = 0

        return lint_message, pep8_grade

    def check_doctests(self, module_path: str) -> tuple[str, bool]:
        """
        Checks whether each function has at least one doctest. Runs all doctests.
        """

        module_name = os.path.splitext(os.path.basename(module_path))[0]
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        functions = inspect.getmembers(module, inspect.isfunction)

        missed_doctests = []
        for name, func in functions:

            finder = doctest.DocTestFinder()
            tests = finder.find(func)

            if len(tests) == 0 or len(tests[0].examples) == 0:
                missed_doctests.append(name)

        result = doctest.testmod(module)

        doctest_message = []
        verdict = True

        if len(missed_doctests) > 0:
            doctest_message.append(f"Following functions miss doctests: {', '.join(missed_doctests)}")
            verdict = False

        if result.failed != 0:
            doctest_message.append(f"TestResults({result.failed=}, {result.attempted=})")
            verdict = False

        return "\n".join(doctest_message), verdict

    def run(self, test_suite: TestSuite) -> GraderTestResult:
        """Run tests and count total grade"""

        test_cases = self._discover_test_cases(test_suite)

        pylint_check = False
        doctests_check = False
        for case in test_cases:
            if hasattr(case, "pylint_check") and case.pylint_check:
                pylint_check = True
            if hasattr(case, "doctests_check") and case.doctests_check:
                doctests_check = True

        self.stream.writeln(f"=== Running tests for: {get_testing_script_path()} ===")

        if os.path.exists(get_testing_script_path()):
            result = super().run(test_suite)
            self.show_points(test_cases, result)
            doctest_dependent_points = result.gained_points * (1 - PYLINT_WEIGHT)
            pylint_dependent_points = result.gained_points - doctest_dependent_points

            if pylint_check:
                lint_message, pep8_grade = self.run_pylint()

                pylint_dependent_points *= pep8_grade

                self.stream.writeln(lint_message)

            if doctests_check:
                doctest_message, verdict = self.check_doctests(get_testing_script_path())
                if not verdict:
                    self.stream.writeln(doctest_message)
                    doctest_dependent_points /= 2

            result.gained_points = pylint_dependent_points + doctest_dependent_points
            self.stream.writeln(f"Grade: {result.gained_points}/{result.total_points}")
        else:
            result = GraderTestResult(stream=self.stream, descriptions="", verbosity=0)
            self.show_points(test_cases, result)
            result.gained_points = 0
            self.stream.writeln(f"File '{get_testing_script_path()}' was not found")

        self.stream.seek(0)
        result.output = self.stream.read()

        return result

    def show_points(self, test_cases: list[Type], result: GraderTestResult) -> None:
        """
        Count and print gained points
        Args:
            test_case: class of testcase, we must pass it because, after running tests
            it is very hard to find TestCase class
            result: Test result of the TestCase
        """

        all_methods = [method
                       for test_case in test_cases
                       for name, method in inspect.getmembers(test_case)
                       if name.startswith("test_")]

        failed_methods_names = [method._testMethodName for method, _ in result.failures] + \
                               [method._testMethodName for method, _ in result.errors]
        succeeded_methods = [method for method in all_methods if method.__name__ not in failed_methods_names]
        total_points = sum(getattr(method, "points_increment") for method in all_methods)
        gained_points = sum(getattr(method, "points_increment") for method in succeeded_methods)
        result.total_points = total_points
        result.gained_points = gained_points

    @staticmethod
    def _discover_test_cases(test_suite: TestSuite) -> set:
        """Get all TestCase classes from a TestSuite
        Args:

            test_suite: suite where to search for TestCase classes
        """
        discovered_test_cases = set()

        nodes_to_discover = [test_suite]
        while nodes_to_discover:
            node = nodes_to_discover.pop()

            for test in node._tests:
                if isinstance(test, TestSuite):
                    nodes_to_discover.append(test)
                else:
                    discovered_test_cases.add(test.__class__)

        return discovered_test_cases


def load_python_tests(test_file_path: str) -> TestSuite:
    """Load test suite from file by path

    Args:
        test_file_path: path to the test script

    """
    directory_path = os.path.dirname(test_file_path)
    test_file_name = os.path.basename(test_file_path)

    loader = TestLoader().discover(start_dir=directory_path, top_level_dir=directory_path,
                                   pattern=test_file_name)
    return loader


def load_tests(test_script_path: str) -> TestSuite:
    """Load test suite from python or .tests file"""

    if test_script_path.endswith(".py"):
        suite = load_python_tests(test_script_path)
    elif test_script_path.endswith(".tests"):
        suite = TestLoader().loadTestsFromTestCase(create_unittest(test_script_path))
    else:
        raise TypeError("Invalid test file extension")

    return suite


def run_tests_setup_argparse(argument_parser: argparse.ArgumentParser) -> None:
    """Add necessary arguments to argparse"""
    argument_parser.add_argument("-t", "--tests", help="Path to the test script")
    argument_parser.add_argument("-s", "--solution", help="Path to the solution")
    argument_parser.add_argument("-p", "--pylint-args", help="Arguments for pylint", required=False, default="")


def run_tests_for(script_path: str, test_script_path: str = None, suite: TestSuite = None,
                  pylint_args=None) -> GraderTestResult:
    """
    Run tests for given script with test_script_path
    :param script_path:
    :param test_script_path:
    :param suite:
    :param pylint_args:
    :return:
    """

    pylint_args = [] if pylint_args is None else pylint_args

    if test_script_path is None and suite is None:
        raise ValueError("You must provide either test_script_path or suite")

    set_testing_script_path(script_path)

    if suite is None:
        suite = load_tests(test_script_path)

    stream = StringIO()

    runner = GraderTestRunner(stream, pylint_args=pylint_args)
    result = runner.run(suite)

    return result


def run_tests(args: argparse.Namespace) -> None:
    """Perform run tests action"""

    test_file_path = args.tests
    solution_file_path = args.solution
    pylint_args = map_pylint_arguments(args.pylint_args)

    result = run_tests_for(script_path=solution_file_path, test_script_path=test_file_path, pylint_args=pylint_args)

    result.stream.seek(0)
    print(result.stream.read())



from __future__ import annotations

import curses

from . import constants
from .model import Model


class Mode:
    NORMAL = "normal"
    INPUT = "input"


FORBIDDEN_KEYSTROKES = [
    curses.KEY_UP,
    curses.KEY_DOWN,
    curses.KEY_LEFT,
    curses.KEY_RIGHT,
    curses.KEY_ENTER,
    curses.KEY_BACKSPACE,
    10,
    13,
]


class Panel:

    def __init__(self, window, storage: Model):
        self.window = window
        self.storage = storage
        self.focused = False
        self.visible = True
        self.rows, self.cols = self.window.getmaxyx()


def is_enter(c: int) -> bool:
    """Check if enter was entered"""
    return c == curses.KEY_ENTER or c == 10 or c == 13


class StudentSidebar(Panel):

    NAME = "student_sidebar"

    def __init__(self, screen, storage):
        super().__init__(screen, storage)

        self.position = 0

        self.width = len(max(self.storage.students_folders, key=len)) + 2
        self.width = max(self.width, len(max(self.storage.problems, key=len)) + 20)

    def draw(self):

        self.window.addstr(0, 1, "STUDENTS:", curses.A_UNDERLINE)

        self.window.vline(0, self.width, curses.ACS_VLINE, 1)
        self.window.vline(1, self.width, curses.ACS_SSSB, 1)
        self.window.vline(2, self.width, curses.ACS_VLINE, self.rows - SummaryPanel.HEIGHT - 2)

        for i, student in enumerate(self.storage.students_folders):
            if i == self.position and self.focused:
                mode = curses.A_REVERSE
            elif i == self.position and not self.focused:
                mode = constants.GREEN_COLOR | curses.A_ITALIC
            else:
                mode = curses.A_NORMAL
            self.window.addstr(i + 1, 1, student, mode)

    def control(self, c):
        if c == curses.KEY_UP and self.position > 0:
            self.position -= 1
        elif c == curses.KEY_DOWN and self.position < len(self.storage.students_folders) - 1:
            self.position += 1
        elif is_enter(c):
            self.storage.set_current_panel(ProblemsTabs.NAME)

        self.storage.current_student = self.storage.students_folders[self.position]


class ProblemsTabs(Panel):
    NAME = 'problem_tabs'

    def __init__(self, window, storage):
        super().__init__(window, storage)

        self.position = 0

        self.problems = [rule.filename for rule in storage.grading_schema.rules]

    def draw(self):

        padding = self.storage.panels[StudentSidebar.NAME].width

        self.window.addstr(0, padding + 2, "FILES:", curses.A_UNDERLINE)
        self.window.hline(1, padding, curses.ACS_HLINE, self.cols - 3)

        padding += 7

        for i, student in enumerate(self.problems):
            if i == self.position and self.focused:
                mode = curses.A_REVERSE
            elif i == self.position and not self.focused:
                mode = constants.GREEN_COLOR | curses.A_ITALIC
            else:
                mode = curses.A_NORMAL

            self.window.addstr(0, padding + 2, student, mode)
            padding += len(student) + 1

    def control(self, c):
        match c:
            case curses.KEY_LEFT if self.position > 0:
                self.position -= 1
            case curses.KEY_RIGHT if self.position < len(self.problems) - 1:
                self.position += 1
            case curses.KEY_ENTER | 10 | 13:
                self.storage.run_test_for_selected_problem()
                self.storage.set_current_panel(TestsResultPanel.NAME)
                self.storage.panels[TestsResultPanel.NAME].reset_cursor()

        self.storage.current_problem_file = self.problems[self.position]


class TestsResultPanel(Panel):
    NAME = 'tests_result_panel'
    TOP_MARGIN = 2

    def __init__(self, window, storage):
        super().__init__(window, storage)

        self.start_y = 0
        self.start_x = 0
        self.margin = 0
        self.mode = Mode.NORMAL

        self.visible_vertical_range = self.rows - 3
        self.visible_horizontal_range = 0

        self.x, self.y = 0, 0

        self.length_of_current_line = None
        self.max_horizontal_scroll = None
        self.maximum_x = None
        self.maximum_y = None

        self.lines = None

    @property
    def current_line(self) -> str:
        """Return current line"""
        return self.lines[self.start_y + self.y]

    @property
    def actual_y(self) -> str:
        return self.y + self.start_y

    @property
    def actual_x(self) -> str:
        return self.x + self.start_x

    def draw(self):

        self.margin = self.storage.panels[StudentSidebar.NAME].width + 2
        self.visible_horizontal_range = self.cols - self.margin

        self.lines = self.storage.get_current_test_output().split('\n')

        for i, line in enumerate(self.lines[self.start_y: self.start_y + self.visible_vertical_range]):
            self.window.addstr(2 + i, self.margin, line[self.start_x:self.start_x + self.visible_horizontal_range])

        if self.focused:
            curses.curs_set(1)
            self.window.move(self.TOP_MARGIN + self.y, self.margin + self.x)
        else:
            curses.curs_set(0)

    def control(self, c):

        self.count_navigation_variables()

        match c:
            case curses.KEY_UP:
                self.navigate_up()
            case curses.KEY_DOWN:
                self.navigate_down()
            case curses.KEY_RIGHT:
                self.navigate_right()
            case curses.KEY_LEFT:
                self.navigate_left()
            case 36 | 59 if self.mode == Mode.NORMAL:  # $
                self.navigate_end_line()
            case 94 | 58 if self.mode == Mode.NORMAL:  # ^
                self.navigate_start_line()
            case 71 | 159 if self.mode == Mode.NORMAL:  # G
                self.start_y = max(0, len(self.lines) - self.maximum_y)
                self.y = self.maximum_y - 1
            case 103 | 191 if self.mode == Mode.NORMAL:  # g
                self.start_y = 0
                self.y = 0
            case 114 | 186 if self.mode == Mode.NORMAL:  # r
                self.storage.run_test_for_selected_problem()

    def count_navigation_variables(self):
        self.length_of_current_line = len(self.lines[self.start_y + self.y])
        self.max_horizontal_scroll = max(0, self.length_of_current_line - self.visible_horizontal_range)
        self.maximum_x = min(self.visible_horizontal_range - 1, self.length_of_current_line)
        self.maximum_y = min(len(self.lines), self.visible_vertical_range)

    def navigate_left(self) -> None:
        if self.x > 0:
            self.x -= 1

        if self.x == 0:
            self.start_x = max(self.start_x - 1, 0)

    def navigate_right(self) -> None:
        if self.x < self.maximum_x:
            self.x += 1

        if self.x == self.maximum_x:
            self.start_x = min(self.start_x + 1, self.max_horizontal_scroll)

    def navigate_up(self) -> None:
        if self.y > 0:
            self.y -= 1

        if self.y == 0:
            self.start_y = max(self.start_y - 1, 0)

    def navigate_down(self) -> None:
        if self.y < self.maximum_y - 1:
            self.y += 1

        if self.y == self.maximum_y - 1:
            self.start_y = min(self.start_y + 1, len(self.lines) - self.maximum_y - 1)

    def navigate_start_line(self) -> None:
        self.x = 0
        self.start_x = 0

    def navigate_end_line(self) -> None:
        self.start_x = max(0, self.length_of_current_line - self.visible_horizontal_range)
        self.x = self.maximum_x

    def reset_cursor(self):
        self.x, self.y = 0, 0


class SummaryPanel(Panel):

    NAME = 'summary_panel'
    HEIGHT = 15

    def __init__(self, window, storage):
        super().__init__(window, storage)

    def draw(self):

        width = self.storage.panels[StudentSidebar.NAME].width
        start_y = self.rows - self.HEIGHT

        self.window.hline(start_y, 0, curses.ACS_HLINE, width)
        self.window.addch(start_y, width, curses.ACS_SBSS)
        self.window.vline(start_y + 1, width, curses.ACS_VLINE, self.HEIGHT)

        self.window.addstr(start_y + 1, 1, "SUMMARY:", curses.A_UNDERLINE)

        total_grade = 0

        for i, rule in enumerate(self.storage.grading_schema.rules):
            test_result = self.storage.test_runner_result[self.storage.current_student].get(rule.filename)

            if test_result is not None:
                gained_grade = test_result.current_grade
                max_grade = test_result.max_grade
                reduction = test_result.reduction
            else:
                max_grade, gained_grade, reduction = 0, 0, 0

            total_grade += gained_grade

            result_string = "No Data" if test_result is None else f"{gained_grade}/{max_grade}"
            result_string = f" - {rule.filename}: {result_string}"

            reduction_string = f" (-{reduction:.1f})" if reduction else ""

            self.window.addstr(start_y + i + 2, 1, result_string)
            self.window.addstr(start_y + i + 2, 1 + len(result_string), reduction_string, constants.RED_COLOR)

        self.window.addstr(start_y + len(self.storage.grading_schema.rules) + 2, 2,
                           f"> Total Grade: {total_grade}/{self.storage.grading_schema.total_grade}")


class EditorPanel(TestsResultPanel):

    NAME = 'editor_panel'

    def __init__(self, window, storage):
        super().__init__(window, storage)


    def draw(self):

        self.margin = self.storage.panels[StudentSidebar.NAME].width + 2
        self.visible_horizontal_range = self.cols - self.margin

        self.lines = self.storage.current_test_result.summary.split('\n') if self.storage.current_test_result is not None else [""]

        self.window.addstr(2, self.margin, f"SUMMARY EDITOR: {self.mode} mode", curses.A_UNDERLINE)
        for i, line in enumerate(self.lines[self.start_y: self.start_y + self.visible_vertical_range]):
            self.window.addstr(3 + i, self.margin, line[self.start_x:self.start_x + self.visible_horizontal_range])

        if self.focused:
            curses.curs_set(1)
            self.window.move(self.TOP_MARGIN + self.y + 1, self.margin + self.x)
        else:
            curses.curs_set(0)

    def control(self, c):
        """Control behaviour of the editor panel."""
        super().control(c)

        match c:
            case 105 if self.mode == Mode.NORMAL:  # i
                self.mode = Mode.INPUT
            case curses.KEY_F1 if self.mode == Mode.INPUT:
                self.mode = Mode.NORMAL
            case ch if self.mode == Mode.INPUT and ch not in FORBIDDEN_KEYSTROKES:
                self.lines[self.actual_y] = self.current_line[:self.actual_x] + chr(ch) + self.current_line[self.actual_x:]
                self.count_navigation_variables()
                self.navigate_right()
            case curses.KEY_BACKSPACE if self.mode == Mode.INPUT and self.actual_x != 0:
                self.lines[self.actual_y] = self.current_line[:self.actual_x - 1] + self.current_line[self.actual_x:]
                self.count_navigation_variables()
                self.navigate_left()
            case curses.KEY_BACKSPACE if self.mode == Mode.INPUT and self.actual_x == 0:
                current_line = self.lines.pop(self.actual_y)
                
            case curses.KEY_ENTER | 10 | 13:
                current_line = self.current_line
                self.lines[self.actual_y] = current_line[:self.actual_x]
                self.lines.insert(self.actual_y + 1, current_line[self.actual_x:])
                self.count_navigation_variables()
                self.navigate_start_line()
                self.navigate_down()

        self.storage.current_test_result.summary = '\n'.join(self.lines)



"""Decorators for testing"""
import inspect
import re
from collections import namedtuple
from functools import wraps
from importlib import util

from .utils import capture_io

Case = namedtuple('Case', ['function', 'output', 'points'])


def add_points(points: float):
    def decorator(test_function: callable):
        setattr(test_function, 'points_increment', points)
        return test_function

    return decorator


def run_solution(inputs: list):
    def decorator(test_function: callable):
        @wraps(test_function)
        def wrapper(class_self):
            from .globals import testing_script_path
            outputs = capture_io(testing_script_path, map(str, inputs))
            outputs = outputs.replace("i", "і")
            return test_function(class_self, outputs)

        return wrapper

    return decorator


def import_solution_module():
    def decorator(test_function: callable):
        @wraps(test_function)
        def wrapper(class_self):
            from .globals import testing_script_path
            spec = util.spec_from_file_location("test.test", testing_script_path)
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return test_function(class_self, module)

        return wrapper

    return decorator


def create_function_for_case(name: str, case: Case) -> callable:
    """
    Create test-function for case
    :param case: case of testing
    :param name: name of test case
    :return: function
    """

    call_code = re.findall(r"\.(.*?)\,", inspect.getsource(case.function))[0]

    @import_solution_module()
    @add_points(case.points)
    def test_case(self, module):
        self.assertEqual(case.function(module), case.output, f"{call_code} should equal to {case.output}")

    test_case.__name__ = name

    return test_case


def resolve_test_cases(cls):
    cases_arrays = {cases_name: cases for cases_name, cases in cls.__dict__.items() if cases_name.endswith("_cases")}

    for case_name, cases in cases_arrays.items():
        for i, case in enumerate(cases):
            name = f"test_{case_name}_{i}"
            setattr(cls, name, create_function_for_case(name=name, case=case))

    return cls


def run_pylint_check(cls):
    setattr(cls, "pylint_check", True)
    return cls


def run_doctests_check(cls):
    setattr(cls, "doctests_check", True)
    return cls



from __future__ import annotations

import os
import shutil
from copy import deepcopy
from dataclasses import dataclass

import pyperclip
from numpy.f2py.auxfuncs import throw_error

from ..classroom.commands import collect_test_scripts, TMP_TESTS_DIR
from ..classroom.model import GradingSchema
from ..test_runner import load_tests, GraderTestResult, run_tests_for

REDUCTION_STEP = 0.3


@dataclass
class TestResult:
    """Class to hold the test result"""
    gained_grade: float
    max_grade: float
    reduction: float = 0
    tester_output: str = ""
    summary: str = ""

    @property
    def current_grade(self) -> float:
        """Return grade with reduction"""
        return round(self.gained_grade - self.reduction, 2)


class Model:
    """Class for storing global data of app"""

    def __init__(self, grading_schema: GradingSchema):
        self.current_panel = None
        self.panels = {}

        self.grading_schema = grading_schema

        self.students_folders = get_students_folders()
        self.problems = [rule.filename for rule in self.grading_schema.rules]

        self.current_student = self.students_folders[0]
        self.current_problem_file = grading_schema.rules[0].filename

        self.test_runner_result = {student: {} for student in self.students_folders}
        self.test_cases = {}

        self.load_test_cases()

    @property
    def current_test_result(self) -> TestResult:
        """Return the current test result"""
        if self.current_problem_file not in self.test_runner_result[self.current_student]:
            return None

        return self.test_runner_result[self.current_student][self.current_problem_file]

    def load_test_cases(self) -> None:
        """Load test cases from specified locations in grading schema and write them to self.test_cases"""

        for rule in self.grading_schema.rules:
            self.test_cases[rule] = load_tests(rule.test_path)

    def add_panel(self, name: str, panel) -> None:
        """Add panel to panel registry"""
        self.panels[name] = panel

    def set_current_panel(self, name: str) -> None:
        """Set current panel"""
        if self.current_panel is not None:
            self.current_panel.focused = False

        self.current_panel = self.panels.get(name)
        self.current_panel.visible = True
        self.current_panel.focused = True

    def get_current_test_output(self) -> str:
        if self.current_problem_file not in self.test_runner_result[self.current_student]:
            return ""

        return self.test_runner_result[self.current_student][self.current_problem_file].tester_output

    def get_current_test_result(self) -> GraderTestResult | None:
        if self.current_problem_file not in self.test_runner_result[self.current_student]:
            return None

        return self.test_runner_result[self.current_student][self.current_problem_file]

    def run_test_for_selected_problem(self) -> None:
        script_path = os.path.join(self.current_student, self.current_problem_file)

        grading_rule = next(filter(lambda r: r.filename == self.current_problem_file, self.grading_schema.rules))

        result = run_tests_for(script_path=script_path, suite=deepcopy(self.test_cases[grading_rule]),
                               pylint_args=self.grading_schema.pylint_args)

        rule = next(filter(lambda r: r.filename == self.current_problem_file, self.grading_schema.rules))
        max_grade = self.grading_schema.total_grade * rule.weight / 100
        gained_grade = round(result.gained_points / result.total_points * max_grade, 2)

        test_result = TestResult(gained_grade=gained_grade, max_grade=max_grade, tester_output=result.output)

        self.test_runner_result[self.current_student][self.current_problem_file] = test_result

    def reduce_grade(self) -> None:
        """
        Reduce gained grade of current problem by REDUCTION_STEP
        """
        if self.current_test_result is None:
            return

        if self.current_test_result.current_grade - REDUCTION_STEP >= 0:
            self.current_test_result.reduction += REDUCTION_STEP

        self.current_test_result.reduction = round(self.current_test_result.reduction, 2)

    def cancel_reduction(self) -> None:
        """
        Cancel the reduction on current problem by one REDUCTION_STEP
        :return: None
        """
        if self.current_test_result is None:
            return

        if self.current_test_result.current_grade + REDUCTION_STEP <= self.current_test_result.gained_grade:
            self.current_test_result.reduction -= REDUCTION_STEP

        self.current_test_result.reduction = round(self.current_test_result.reduction, 2)

    def copy_summary_to_clipboard(self) -> None:
        """
        Copy generated summary to clipboard
        :return:
        """
        summary_str_parts = ["SUMMARY:"]

        total_grade = 0

        for i, rule in enumerate(self.grading_schema.rules):
            test_result = self.test_runner_result[self.current_student].get(rule.filename)

            if test_result is not None:
                gained_grade = test_result.current_grade
                max_grade = test_result.max_grade
            else:
                max_grade, gained_grade, reduction = 0, 0, 0

            total_grade += gained_grade

            result_string = "No Data" if test_result is None else f"{gained_grade}/{max_grade}"
            result_string = f" - {rule.filename}: {result_string}"
            summary_str_parts.append(result_string)

        summary_str_parts.append(f"> Total Grade: {total_grade}/{self.grading_schema.total_grade}")

        pyperclip.copy('\n'.join(summary_str_parts))


def get_students_folders() -> list[str]:
    """Return a list of student folders from current directory"""
    return [path for path in os.listdir() if os.path.isdir(path) and
            os.path.exists(os.path.join(path, '.submission.yaml'))]



"""Interactive grading tool"""
from __future__ import annotations
import curses
import traceback
import argparse

from . import constants
from .model import Model
from .panels import StudentSidebar, ProblemsTabs, TestsResultPanel, SummaryPanel, EditorPanel
from ..classroom.commands import parse_grading_schema


def run_tests_interactive_argument_parser(argument_parser: argparse.ArgumentParser) -> None:
    """Add necessary arguments for interactive grading"""
    argument_parser.add_argument('-ap', '--assets-path', required=False)


def create_color_pairs() -> None:
    """Init color pairs"""
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_RED, curses.COLOR_BLACK)

    constants.GREEN_COLOR = curses.color_pair(1)
    constants.RED_COLOR = curses.color_pair(2)


def wrapper(screen, args: argparse.Namespace) -> None:

    create_color_pairs()
    storage = Model(parse_grading_schema(args.grading_schema, assets_path=args.assets_path))

    sidebar = StudentSidebar(screen, storage)
    problem_tabs = ProblemsTabs(screen, storage)
    summary_panel = SummaryPanel(screen, storage)
    test_result_panel = TestsResultPanel(screen, storage)
    editor_panel = EditorPanel(screen, storage)

    # Must be in such order to maintain proper drawing
    storage.add_panel(ProblemsTabs.NAME, problem_tabs)
    storage.add_panel(SummaryPanel.NAME, summary_panel)
    storage.add_panel(StudentSidebar.NAME, sidebar)
    storage.add_panel(TestsResultPanel.NAME, test_result_panel)

    storage.set_current_panel(StudentSidebar.NAME)

    sidebar.draw()
    problem_tabs.draw()
    summary_panel.draw()

    while True:
        screen.erase()

        for panel in storage.panels.values():
            if panel.visible:
                panel.draw()

        c = screen.getch()
        match c:
            case 113:
                break
            case 97 | 132:
                storage.set_current_panel(StudentSidebar.NAME)
            case 115 | 150:
                storage.set_current_panel(ProblemsTabs.NAME)
            case 100 | 178:
                editor_panel.visible = False
                storage.set_current_panel(TestsResultPanel.NAME)
            case 44 | 177:  # , make reduction
                storage.reduce_grade()
            case 46 | 142:  # . - erase reduction
                storage.cancel_reduction()
            case 121 | 189:
                storage.copy_summary_to_clipboard()

        storage.current_panel.control(c)


def run_tests_interactive(args: argparse.Namespace) -> None:
    """Run the test with terminal interface"""
    try:
        curses.wrapper(wrapper, args)
    except Exception as ex:
        traceback.print_exc()
